"""Module."""

from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

import datetime
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from typing import Any

from pandas import to_datetime

from gscipy.csv_io import CSVReader
from gscipy.general import TypeCheckedList
from gscipy.json_io import CustomJSONDict, JsonIOInterface


class Time(JsonIOInterface):
    """Represents a time.

    A Time instance can be initialized by parsing:

    1) a datetime instance
    >>> from datetime import datetime
    >>> t = datetime.now()
    >>> t
    datetime.datetime(2022, 11, 24, 15, 56, 31, 591748)
    >>> Time(t)
    Time(1669301791.591748)

    2) a timestamp (float or int)
    >>> Time(t.timestamp())
    Time(1669301791.591748)
    >>> Time(123)
    Time(123)

    3) a timestamp-convertible string with the correct format
    >>> Time('2022-11-24 15:56:31.591748')
    Time(1669305391.591748)

    raise a TypeError otherwise.

    A Time instance can be modified using a property setter
    >>> t = Time(1)
    >>> t
    Time(1)
    >>> t.time = 2
    >>> t
    Time(2)

    Magic methods for addition, subtraction, multiplication, division, rising to the power, abs()
    and comparison (<, <=, ==, !=, >=, >) of Time instances or a Time instance with a constant (int, float)
    are implemented.
    """

    # __slots__ = ["time", "_time_str"]

    # Default string format for string parsing to a timestamp
    _time_str = "%Y-%m-%d %H:%M:%S.%f%Z"

    def __init__(self, value: datetime | float | int | str = 0) -> None:
        """
        ---

            value : (datetime, float, int, str), optional, default to 0
        """
        if not isinstance(value, (datetime, float, int, str)):
            raise TypeError(
                f"{self.__class__.__name__}() argument must be a timestamp ('float' or 'int'), a datetime"
                f" that can be converted to one, or a string of the format {self.__class__._time_str}"
                f" that can be parsed into one, not '{value.__class__.__name__}'"
            )
        self._time = value
        self.time = self._time

    @classmethod
    def from_json(cls, json_dict: CustomJSONDict) -> Time:
        return cls(json_dict["time"])

    def to_json(self) -> CustomJSONDict:
        json_dict = CustomJSONDict({"time": self.time})
        json_dict["__class__"] = self._get__class__()
        return json_dict

    @property
    def time(self) -> int | float:
        """Get time property.

        >>> t = Time(1)
        >>> t.time
        1
        """
        return self._time

    @time.setter
    def time(self, value: float | int | str | datetime) -> None:
        """Set time property from a datetime instance, a timestamp (float, int) or a timestamp-convertible string.

        >>> t = Time(1)
        >>> t.time = 2
        >>> t.time
        2
        >>> t
        Time(2)

        Parameters
        ----------
         value : (float, int, str, datetime), required
        """
        if not isinstance(value, (datetime, float, int, str)):
            raise TypeError(
                f"{self.__class__.__name__}() argument must be a timestamp ('float' or 'int'), a datetime"
                f" that can be converted to one, or a string of the format {self.__class__._time_str}"
                f" that can be parsed into one, not '{value.__class__.__name__}'"
            )

        if isinstance(value, datetime):
            try:
                value = value.timestamp()
            except Exception as error:
                raise ValueError(
                    f"{self.__class__.__name__}() could not create a valid timestamp from 'value'"
                ) from error

        elif isinstance(value, str):
            try:
                value = to_datetime(value).timestamp()
            except Exception as error:
                raise ValueError(
                    f"{self.__class__.__name__}() could not parse a valid timestamp "
                    f"from 'value' \n (provide a string with format: {self._time_str})"
                ) from error

        self._time = value

    def __call__(self, value: datetime | float | int | str):
        """Make Time instance callable.

        >>> t = Time()
        >>> t
        Time(0)
        >>> t1 = t(2)
        >>> t1
        Time(2)
        """
        return self.__class__(value)

    def __add__(self, other: Time | int | float) -> Time:
        """Add Times or constant to Time, raise TypeError otherwise.

        >>> t1 = Time(1)
        >>> t2 = Time(3)
        >>> t1 + t2
        Time(4)
        >>> t1 + 4
        Time(5)
        """
        if isinstance(other, Time):
            return Time(self.time + other.time)
        if isinstance(other, (int, float)):
            return Time(self.time + other)
        return NotImplemented

    def __radd__(self, other: Time | int | float) -> Time:
        return self + other

    def __sub__(self, other: Time | int | float) -> Time:
        """Subtracts Times or constant to Time, raise TypeError otherwise.

        >>> t1 = Time(1)
        >>> t2 = Time(3)
        >>> t1 - t2
        Time(-2)
        >>> t1 - 4
        Time(-3)
        """
        if isinstance(other, (Time, int, float)):
            return self + (-other)
        return NotImplemented

    def __rsub__(self, other: Time | int | float) -> Time:
        return other + (-self)

    def __mul__(self, other: Time | int | float) -> Time:
        """Multiply Times or constant to Time, raise TypeError otherwise.

        >>> t1 = Time(1)
        >>> t2 = Time(3)
        >>> t1 * t2
        Time(3)
        >>> t1 * 4
        Time(4)
        """
        if isinstance(other, Time):
            return Time(self.time * other.time)
        if isinstance(other, (int, float)):
            return Time(self.time * other)
        return NotImplemented

    def __rmul__(self, other: Time | int | float) -> Time:
        return self * other

    def __truediv__(self, other) -> Time:
        """Divide Times or constant to Time, raise TypeError otherwise.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 / t2
        Time(0.5)
        >>> t1 / 3
        Time(0.3333333333333333)
        """
        return self * (other**-1)

    def __rtruediv__(self, other) -> Time:
        return other / self

    def __pow__(self, power: int | float) -> Time:
        """Rise Time to the power, raise TypeError otherwise.

        >>> t = Time(4)
        >>> t ** 2
        Time(16)
        >>> t ** 0.5
        Time(2.0)
        """
        if isinstance(power, (int, float)):
            return Time(self.time**power)
        return NotImplemented

    def __neg__(self) -> Time:
        return Time(-self.time)

    def __abs__(self) -> Time:
        """Time absolute value.

        >>> t = Time(-1)
        >>> abs(t)
        Time(1)
        """
        return Time(abs(self.time))

    def __lt__(self, other: Time | int | float) -> bool:
        """Compares Time instances.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 < t2
        True
        >>> t1 < 0
        False
        """
        if isinstance(other, Time):
            return self.time < other.time
        if isinstance(other, (int, float)):
            return self.time < other
        return NotImplemented

    def __le__(self, other: Time | int | float) -> bool:
        """Compares Time instances.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 <= t2
        True
        >>> t1 <= 1
        True
        """
        if isinstance(other, Time):
            return self.time <= other.time
        if isinstance(other, (int, float)):
            return self.time <= other
        return NotImplemented

    def __eq__(self, other: Time | int | float) -> bool:
        """Check equality between Time(s) or with respect to a constant.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 == t2
        False
        >>> t1 == 1
        True
        """
        if isinstance(other, Time):
            return self.time == other.time
        if isinstance(other, (int, float)):
            return self.time == other
        return NotImplemented

    def __ne__(self, other: Time | int | float) -> bool:
        """Check inequality between Time(s) or with respect to a constant.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 != t2
        True
        >>> t1 != 1
        False
        """
        return not self == other

    def __gt__(self, other: Time | int | float) -> bool:
        """Compares Time instances.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 > t2
        False
        >>> t1 < 0
        False
        """
        return not self < other

    def __ge__(self, other: Time | int | float) -> bool:
        """Compares Time instances.

        >>> t1 = Time(1)
        >>> t2 = Time(2)
        >>> t1 >= t2
        False
        >>> t1 >= 1
        True
        """
        return not self <= other

    def __hash__(self) -> int:
        """Return hash value of the given instance."""
        return hash(self.time)

    def __bool__(self) -> bool:
        """Check if Time is null."""
        return Time.time == 0

    def __repr__(self) -> str:
        """Returns a representation of a Time object.

        >>> t = Time(1)
        >>> t
        Time(1)
        """
        return f"{self.__class__.__name__}({self.time})"

    def __str__(self) -> str:
        """Returns a string representation of a Time object.

        >>> t = Time(1)
        >>> print(t)
        Time(1)
        """
        return f"{self.__class__.__name__}({self.time})"


class TimeList(TypeCheckedList):
    """Type checked list of Time instances.

    >>> times = TimeList([])
    >>> times
    TimeList([])

    Implement a load() method for loading times from an iterator (for example from a .csv file row iterator).
    """

    def __init__(self, iterator_arg: list[Time] = None) -> None:
        """
        ---

        iterator_arg : List[Time], optional, default to None
            List of Time instances
        """
        # Initialize the parent class to type check for Time instances
        super().__init__(iterator_arg=iterator_arg, instance_type=Time)

    @classmethod
    def from_json(cls, json_dict: CustomJSONDict) -> TimeList:

        return cls(json_dict["data"])

    def load(self, row_iter: Iterable[dict[str, Any]]) -> None:
        """Load Time instances into the type checked list.

        Parameters
        ----------
        row_iter : Iterator[dict[str, Any]]
            Iterator generating a dictionary dict[str, Any] with key {'time'}.
            For example, the row-iterator of a .csv file.

        Returns
        -------
            A type checked list populated with Vector vectors
        """

        for row in row_iter:

            self.append(Time(row["time"]))

        return None


class TimeListLoader:
    """Loads times into a type-checked list given the column name corresponding to the time values,.

     as stored in the input file.

    >>> source = Path.cwd() / "data" / "sub1.csv"
    >>> usecols = ['time']
    >>> times_loader = TimeListLoader(source, usecols)
    >>> times = times_loader.load()
    >>> times[:3]
    TimeList([Time(1651585280.77981),
             Time(1651585280.78481),
             Time(1651585280.78981)])

    The time, as given by the column name, is automatically re-named to pre-defined name: 'time'

    >>> times[0].time
    1651585280.77981
    """

    # Pre-defined time name
    new_colnames = ["time"]

    def __init__(self, source: Path, usecols: list[str]) -> None:
        """
        ---

            source : Path, required
                Path of the input file.

            usecols : List or Tuple, required
                Names corresponding to the time values as stored in the input file.
        """
        self.source = source
        self.usecols = usecols

        # Map given time name to pre-defined name
        self.col_map_dict = dict(zip(self.usecols, self.new_colnames))

    def load(self) -> TimeList:
        """Loads times into the type checked list from .csv file.

        Returns
        -------
            A type checked list populated with times.
        """
        times = TimeList()
        csv_reader = CSVReader(self.source)
        row_iter = csv_reader.row_iter(
            usecols=self.usecols, col_map_dict=self.col_map_dict
        )
        times.load(row_iter)

        return times



"""Module for handling data input/output from and to files.
"""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

import csv
from pathlib import Path
from typing import Iterator, Iterable

import pandas as pd

# TODO: write fun to create columns mapping dictionary def create_columns_mapping_dict(usecols, new_colnames):
#  return dict(zip(usecols, new_colnames))


class CSVReader:
    """
    Class to read .csv files.

    Attributes
    ----------
        source : Path, required
            Path of a .csv file.

    Methods
    -------
        to_data_frame
            Read .csv file into a pd.DataFrame.

        row_iter
            Iterate over the rows of the .csv file mapped into a dictionary with header column names as key values.
    """
    def __init__(self, source: Path) -> None:
        """
            source : Path, required
                Path of a .csv file.
        """
        self.source = source

    @staticmethod
    def _rename_keys(
        old_dict: dict[str, str], key_map_dict: dict[str, str]
    ) -> dict[str, str]:
        """
        Rename keys of input dictionary according to mapping dictionary.

        Parameters
        ----------
        old_dict : dict(str, str), required

        key_map_dict : dict[str, str], required
            Dictionary mapping old to new keys -> {'old_key_1': 'new_key_1, 'old_key_2': 'new_key_2',}

        Returns
        -------
            The input dictionary with the new keys.
        """
        return {
            (key_map_dict[key] if key in key_map_dict else key): value
            for key, value in old_dict.items()
        }

    @staticmethod
    def _filter_by_key(old_dict: dict, keys: Iterable[str]) -> dict:
        """
        Filter dictionary by key value.

        Discard items of old_dict not in the given list of keys.

        Parameters
        ----------
        old_dict : dict, required

        keys : List or Tuple, required
            Discard dictionary items not in keys.

        Returns
        -------
            A subset of the input dictionary.
        """
        return {key: value for key, value in old_dict.items() if key in keys}

    @staticmethod
    def _check_missing_columns(
        csv_reader: csv.DictReader, usecols: Iterable[str]
    ) -> None:
        """
        Check if selected column names exists in the .csv file, raise ValueError otherwise.

        Parameters
        ----------
        csv_reader : csv.DictReader instance, required
            Instance of csv.DictReader.

        usecols : List or Tuple, required
            List of column names.

        Returns
        -------
            None if all column names exists in the .csv file header, raise ValueError otherwise.
        """
        header = csv_reader.fieldnames
        missing_cols = [col for col in usecols if col not in header]

        if missing_cols:
            raise ValueError(f"invalid column name(s): {missing_cols}")

    def row_iter(
        self,
        *,
        usecols: Iterable[str] = None,
        col_map_dict: dict[str, str] = None,
    ) -> Iterator[dict[str, str]]:
        """
        Iterator over the rows of the .csv file mapped to a dictionary of strings:

         {'col1_name': 'row_i1', 'col2_name': 'row_i2', 'col3_name': 'row_i3'}.
          Optionally, a subset of columns can be loaded and/or renamed to user defined column names.

        Parameters
        ----------
            usecols : List or Tuple, optional, key-word only, defaults to None
                Returns a subset of the columns. If empty, returns all columns. Raise ValueError for incorrect
                column names.

            col_map_dict : dict[str, str], optional, key-word only, default to None
                Dictionary mapping old (keys) to new (values) column names.

        Returns
        -------
            Iterator over the .csv rows.
        """
        # TODO: decide if ordering the dictionary according to usecols and colmap order
        # ordered_dict = {k : new_colnames_row[k] for k in new_cols}

        with open(self.source, "r") as data:

            # Instantiate csv_reader
            csv_reader = csv.DictReader(data)

            # For each row of the .csv file generate a dictionary: {col_name: value}
            for row in csv_reader:

                if not usecols and not col_map_dict:
                    yield row

                elif not usecols and col_map_dict:
                    yield self._rename_keys(row, col_map_dict)

                elif usecols and not col_map_dict:
                    self._check_missing_columns(csv_reader, usecols)
                    yield self._filter_by_key(row, usecols)

                elif usecols and col_map_dict:
                    self._check_missing_columns(csv_reader, usecols)
                    filtered_row = self._filter_by_key(row, usecols)
                    yield self._rename_keys(filtered_row, col_map_dict)

    def to_data_frame(self, **kwargs) -> pd.DataFrame:
        """
        Read .csv file into a pd.DataFrame.

        Parameters
        ----------
            **kwargs : dict, required
                keyword arguments of pandas.read_csv() method.
                See: https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html

        Returns
        -------
            pd.DataFrame
        """
        return pd.read_csv(self.source, **kwargs)



"""Module."""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Any

import numpy as np

from gscipy.csv_io import CSVReader
from gscipy.decorators import check_args_len
from gscipy.general import TypeCheckedList
from gscipy.json_io import CustomJSONDict, JsonIOInterface


class VectorNDim:
    """Represents a vector in a multi-dimensional space.

    A vector instance is instantiated by key-word only arguments.

    >>> v = VectorNDim(x0=1, x1=1, x2=1, x3=1)

    Each coordinate is assigned to a class attribute
    >>> v.x0
    1.0
    >>> v.x1
    1.0
    >>> v.x2
    1.0
    >>> v.x3
    1.0

    The following operations are implemented by operator overloading:

    Get the total number of coordinates (vector length)
    >>> len(v)
    4

    Absolute value (magnitude)
    >>> v.magnitude
    2.0
    >>> abs(v)
    2.0

    Add vector inplace
    >>> v = VectorNDim(x0=1, x1=1, x2=1, x3=1)
    >>> v
    Vector(1.0, 1.0, 1.0, 1.0)
    >>> v += v
    >>> v
    Vector(2.0, 2.0, 2.0, 2.0)
    >>> v += v
    >>> v
    Vector(4.0, 4.0, 4.0, 4.0)

    Get coordinate by index
    >>> v = VectorNDim(x0=1, x1=1, x2=1, x3=1)
    >>> v[1]
    1.0

    Set coordinate by index
    >>> v[1] = 10
    >>> v
    Vector(1.0, 10, 1.0, 1.0)

    Iterate over the vector coordinates
    >>> [c for c in v]
    [1.0, 10, 1.0, 1.0]

    Dot product:
    >>> v1 = VectorNDim(x0=1, x1=1, x2=1, x3=1)
    >>> v2 = VectorNDim(x0=2, x1=2, x2=2, x3=2)

    >>> v1.dot(v2)
    8.0
    >>> v1 @ v2
    8.0

    Vector addition
    >>> v1 + v2
    Vector(3.0, 3.0, 3.0, 3.0)

    Vector substraction
    >>> v1 - v2
    Vector(-1.0, -1.0, -1.0, -1.0)

    Coordinate-wise and scalar multiplication
    >>> v1 * v2
    Vector(2.0, 2.0, 2.0, 2.0)

    Multiplication by a constant
    >>> v1 * 5
    Vector(5.0, 5.0, 5.0, 5.0)

    Inversion
    >>> -v1
    Vector(-1.0, -1.0, -1.0, -1.0)

    Raising to the power of a number (int or float)
    >>> v2 ** 2
    Vector(4.0, 4.0, 4.0, 4.0)

    Vector comparison
    >>> v1 = VectorNDim(x0=1, x1=1, x2=1, x3=1)
    >>> v2 = VectorNDim(x0=2, x1=2, x2=2, x3=2)
    >>> v1 == v1
    True
    >>> v1 == v2
    False
    >>> v1 != v2
    True

    See: See: https://gist.github.com/gmalmquist/6b0ee1c38d7263f6158f
    """

    def __init__(self, **coordinates: float | int | str) -> None:

        self._class_name = self.__class__.__name__

        # Set keyword arguments to attributes
        for key, value in coordinates.items():
            try:
                setattr(self, key, float(value))
            except ValueError as error:
                raise TypeError(
                    f"{self._class_name}() argument must be a list of float(s) or int(s),"
                    f" or float-convertible string numbers, not '{value.__class__.__name__}'"
                ) from error

        # Set coordinates attribute
        self._coordinates = list(map(float, coordinates.values()))
        self._keys = list(coordinates.keys())

    @property
    def magnitude(self) -> float:
        """Vector magnitude (absolute value).

        >>> v = VectorNDim(x0=1, x1=1)
        >>> v.magnitude
        1.4142135623730951
        """
        return sum(self**2) ** 0.5

    @check_args_len
    def dot(self, other: VectorNDim) -> float:
        """Dot product (inner product) between vectors of same dimension.

        >>> v = VectorNDim(x0=1, x1=1)
        >>> v.dot(v)
        2.0

        raise TypeError if 'other' is not a vector.
        """
        if isinstance(other, self.__class__):
            return sum(self * other)
        raise TypeError(
            f"unsupported operand type(s) for {self._class_name}.dot():"
            f" '{other.__class__.__name__}'"
        )

    def __call__(self, **coordinates: float) -> VectorNDim:
        """Make Vector instances callable.

        >>> v = VectorNDim()
        >>> v
        Vector()
        >>> v1 = v(x1=1, x2=1)
        >>> v1
        Vector(1.0, 1.0)
        """
        return self.__class__(**coordinates)

    def __getitem__(self, index: int) -> float:
        """Get vector coordinate by index or slicing.

        >>> v = VectorNDim(x0=1, x1=2)
        >>> v[0]
        1.0
        """
        return self._coordinates[index]

    def __setitem__(self, index: int, value: int | float) -> None:
        """Set vector coordinate by index.

        >>> v = VectorNDim(x0=1, x1=2)
        >>> v
        Vector(1.0, 2.0)
        >>> v[1] = 123
        >>> v
        Vector(1.0, 123)

        Raise TypeError if 'value' is not a float or int number.
        """
        if isinstance(value, (int, float)):
            self._coordinates[index] = value
            # Update value of all attributes derived from self._coordinates
            unprotected_keys = [
                key for key in self.__dict__.keys() if not key.startswith("_")
            ]
            for i, key in enumerate(unprotected_keys):
                setattr(self, key, self._coordinates[i])
        else:
            raise TypeError(
                f"unsupported operand type(s) for {self._class_name}.__setitem__():"
                f" '{value.__class__.__name__}'"
            )

    def __iter__(self) -> Iterator[float]:
        """Allows iteration over the vector coordinates.

        >>> v = VectorNDim(x1=1, x2=2)
        >>> [c for c in v]
        [1.0, 2.0]
        """
        return iter(self._coordinates)

    def __len__(self) -> int:
        """Returns vector length (number of coordinates).

        >>> v = VectorNDim(x1=1, x2=2)
        >>> len(v)
        2
        """
        return len(self._coordinates)

    @check_args_len
    def __add__(self, other: VectorNDim) -> VectorNDim:
        """Vector addition.

        >>> v1 = VectorNDim(x0=1, x1=1)
        >>> v2 = VectorNDim(x0=2, x1=2)
        >>> v1 + v2
        Vector(3.0, 3.0)

        Raise AssertionError if the two vectors have different dimension.
        Raise TypeError if 'other' is not a vector.
        """
        if isinstance(other, self.__class__):
            return self(
                **{
                    self._keys[i]: c1 + c2
                    for i, (c1, c2) in enumerate(zip(self, other))
                }
            )
        return NotImplemented

    def __radd__(self, other: VectorNDim) -> VectorNDim:
        return self + other

    @check_args_len
    def __iadd__(self, other: VectorNDim) -> VectorNDim:
        """Inplace vector addition.

        >>> v = VectorNDim(x1=1, x2=1)
        >>> v += v
        >>> v
        Vector(2.0, 2.0)

        Raise AssertionError if the two vectors have different dimension.
        Raise TypeError if 'other' is not a vector.
        """
        if isinstance(other, self.__class__):
            return self(
                **{
                    self._keys[i]: c1 + c2
                    for i, (c1, c2) in enumerate(zip(self, other))
                }
            )
        return NotImplemented

    @check_args_len
    def __sub__(self, other: VectorNDim) -> VectorNDim:
        """Vector subtraction.

        >>> v1 = VectorNDim(x0=1, x1=1)
        >>> v2 = VectorNDim(x0=2, x1=2)
        >>> v1 - v2
        Vector(-1.0, -1.0)

        Raise AssertionError if the two vectors have different dimension.
        Raise TypeError if 'other' is not a vector.
        """
        if isinstance(other, self.__class__):
            return self + (-other)
        return NotImplemented

    def __rsub__(self, other: VectorNDim) -> VectorNDim:
        return other + (-self)

    @check_args_len
    def __mul__(self, other: VectorNDim | int | float) -> VectorNDim:
        """Coordinate-wise vector multiplication and scalar multiplication of a vector.

        >>> v1 = VectorNDim(x0=1, x1=1)
        >>> v2 = VectorNDim(x0=2, x1=2)
        >>> v1 * v2
        Vector(2.0, 2.0)
        >>> v1 * 10
        Vector(10.0, 10.0)

        Raise AssertionError if the two vectors have different dimension.
        Raise TypeError if 'other' is not a Vector instance or a number (int, float).
        """
        if isinstance(other, self.__class__):
            return self(
                **{
                    self._keys[i]: a * b
                    for i, (a, b) in enumerate(zip(self, other))
                }
            )
        if isinstance(other, (int, float)):
            return self(
                **{self._keys[i]: a * other for i, a in enumerate(self)}
            )
        return NotImplemented

    def __rmul__(self, other: VectorNDim | int | float) -> VectorNDim:
        return self * other

    def __matmul__(self, other: VectorNDim) -> float:
        """Overload @ operator with dot product."""
        return self.dot(other)

    def __pow__(self, power: int | float) -> VectorNDim:
        """Vector raise to the power.

        >>> v = VectorNDim(x0=4, x1=4)
        >>> v ** 2
        Vector(16.0, 16.0)
        >>> v ** 0.5
        Vector(2.0, 2.0)

        Raise TypeError if 'power' is not a number of type int or float.
        """
        if isinstance(power, (int, float)):
            return self(
                **{self._keys[i]: c**power for i, c in enumerate(self)}
            )
        return NotImplemented

    def __neg__(self) -> VectorNDim:
        """Inverse vector.

        >>> v = VectorNDim(x1=1, x2=1)
        >>> -v
        Vector(-1.0, -1.0)
        """
        return self(**{self._keys[i]: -c for i, c in enumerate(self)})

    def __pos__(self) -> VectorNDim:
        return self

    def __abs__(self) -> float:
        """Absolute value (vector magnitude).

        >>> v = VectorNDim(x0=1, x1=1)
        >>> abs(v)
        1.4142135623730951
        """
        return self.magnitude

    @check_args_len
    def __eq__(self, other: VectorNDim | int | float) -> bool:
        """Check vector equality in terms of both magnitude and direction.

        >>> v1 = VectorNDim(x0=1, x1=2)
        >>> v2 = VectorNDim(x0=2, x1=3)
        >>> v1 == v2
        False
        >>> v1 == v1
        True

        Raise AssertionError if the two vectors have different dimensions.
        """
        if isinstance(other, self.__class__):
            return all(
                a == b for a, b in zip(self._coordinates, other._coordinates)
            )
        return NotImplemented

    def __ne__(self, other: VectorNDim | int | float) -> bool:
        """Check vector vector inequality in terms of both magnitude and direction.

        >>> v1 = VectorNDim(x0=1, x1=2)
        >>> v2 = VectorNDim(x0=2, x1=3)
        >>> v1 != v2
        True
        >>> v1 != v1
        False
        """
        return not self == other

    def __hash__(self) -> int:
        """Return hash value of the given instance."""
        return hash(tuple(self))

    def __bool__(self) -> bool:
        """Check that all coordinates are null."""
        return all(v == 0 for v in self)

    def __repr__(self) -> str:
        """Returns a representation of a Vector object.

        >>> v = VectorNDim(x1=1, x2=3)
        >>> v
        Vector(1.0, 3.0)
        """
        return f"{self._class_name}({', '.join(str(c) for c in self._coordinates)})"

    def __str__(self) -> str:
        """Returns a string representation of a Vector object.

        >>> v = VectorNDim(x1=1, x2=3)
        >>> print(v)
        Vector(1.0, 3.0)
        """
        return f"{self._class_name}({', '.join(str(c) for c in self._coordinates)})"


class Vector(VectorNDim, JsonIOInterface):
    """Represents a vector in 3-dimensions with coordinates (x, y, z).

    >>> r = Vector(x=1, y=2, z=3)
    >>> r
    Position(1, 2, 3)

    Coordinates are set to zero by default
    >>> r = Vector(x=1, y=2)
    >>> r
    Position(1, 2, 0)

    Implement cross-product
    >>> r1 = Vector(x=1, y=2, z=2)
    >>> r2 = Vector(x1=2, y=3)
    >>> r1 ^ r2
    Position(-6.0, 4.0, -1.0)
    >>> r1.cross(r1)
    Position(-6.0, 4.0, -1.0)
    """

    def __init__(self, *, x: float = 0, y: float = 0, z: float = 0) -> None:
        """
        X : float, optional, default to 0.

        y : float, optional, default to 0
        z : float, optional, default to 0
        """
        super().__init__(x=x, y=y, z=z)

    # __slots__ = ["x", "y", "z"]

    @classmethod
    def from_json(cls, json_dict: CustomJSONDict) -> Vector:
        return cls(x=json_dict["x"], y=json_dict["y"], z=json_dict["z"])

    def to_json(self) -> CustomJSONDict:
        obj_dict = {"x": self.x, "y": self.y, "z": self.z}
        json_dict = CustomJSONDict(obj_dict)
        json_dict["__class__"] = self._get__class__()
        return json_dict

    def cross(self, other: Vector) -> Vector:
        """VectorNDim cross product.

        >>> r1 = Vector(x=1, y=2, z=2)
        >>> r2 = Vector(x1=2, y=3)
        >>> r1.cross(r1)
        Position(-6.0, 4.0, -1.0)

        Raise TypeError if 'other' is not a vector.
        """
        if isinstance(other, self.__class__):
            return self ^ other
        raise TypeError(
            f"unsupported operand type(s) for {self._class_name}.cross():"
            f" '{other.__class__.__name__}'"
        )

    def __xor__(self, other: Vector) -> Vector:
        """VectorNDim cross product.

        >>> r1 = Vector(x=1, y=2, z=2)
        >>> r2 = Vector(x1=2, y=3)
        >>> r1 ^ r2
        Position(-6.0, 4.0, -1.0)
        """
        if isinstance(other, self.__class__):
            return self(
                **{
                    self._keys[i]: c
                    for i, c in enumerate(np.cross(self, other))
                }
            )
        return NotImplemented


class VectorsList(TypeCheckedList):
    """Type checked list of 3D vectors.

    >>> vectors = VectorsList([])
    >>> vectors
    VectorsList([])

    Implement a load() method for loading vectors from an iterator (for example from a .csv file row iterator).
    Raise TypeError if a not Vector instance is appended or inserted (in)to the list.
    """

    def __init__(self, iterator_arg: list[Vector] = None):
        """
        Iterator_arg : List[Vector], optional, default to None.

            List of Vector vectors
        """
        # Initialize the parent class to type check for Vector instances
        super().__init__(iterator_arg=iterator_arg, instance_type=Vector)

    @classmethod
    def from_json(cls, json_dict: CustomJSONDict) -> VectorsList:

        return cls(json_dict["data"])

    def load(self, row_iter: Iterable[dict[str, Any]]) -> None:
        """Load 3D vectors into the type checked list.

        Parameters
        ----------
        row_iter : Iterable[dict[str, Any]]
            Iterable generating a dictionary dict[str, Any] with keys {'x', 'y', 'z'}.
            For example, the row-iterator of a .csv file.

        Returns
        -------
            A type checked list populated with Vector vectors
        """
        for row in row_iter:

            self.append(Vector(**row))


class VectorListLoader:
    """Loads 3D vectors into a type-checked list given the column names corresponding to the 3D coordinates values.

    (x, y, z), or a subset of those, as stored in the input file.

    >>> source = Path.cwd() / "data" / "sub1.csv"
    >>> usecols = ['gaze_deg_coord_x', 'gaze_deg_coord_y']
    >>> vectors_loader = VectorListLoader(source, usecols)
    >>> vectors = vectors_loader.load()
    >>> vectors[:3]
    VectorsList([Vector(0.4622481497169438, 3.76608427632374, 0.0),
             Vector(0.46393828577858687, 3.7411490072433953, 0.0),
             Vector(0.456283375837779, 3.7295640810489856, 0.0)])

    The coordinates, as given by the column names, are automatically re-named to pre-defined names: 'x', 'y', 'z'

    >>> vectors[0].__dict__
    {'_class_name': 'Vector',
     'x': 0.4622481497169438,
     'y': 3.76608427632374,
     'z': 0.0,
     '_coordinates': [0.4622481497169438, 3.76608427632374, 0.0],
     '_keys': ['x', 'y', 'z']}
    """

    # Pre-defined 3D coordinates names
    new_names = ["x", "y", "z"]

    def __init__(self, source: Path, usecols: list[str]) -> None:
        """
        ---

            source : Path, required
                Path of the input file.

            usecols : List or Tuple, required
                Names corresponding to the coordinates (x, y, z) of a 3D vector, or a subset of those,
                as stored in the input file.
        """
        self.source = source
        self.usecols = usecols

        # Map given coordinates names to pre-defined names
        self.col_map_dict = dict(zip(self.usecols, self.new_names))

    def load(self) -> VectorsList:
        """Loads 3D vectors into the type checked list from .csv file.

        Returns
        -------
            A type checked list populated with 3D vectors.
        """
        vectors = VectorsList()
        csv_reader = CSVReader(self.source)
        row_iter = csv_reader.row_iter(
            usecols=self.usecols, col_map_dict=self.col_map_dict
        )
        vectors.load(row_iter)

        return vectors



"""Module with operators transforming trajectories objects."""
from __future__ import annotations

__author__ = "Matteo Gabba"
__copyright__ = "Copyright 2022, all right reserved Gabba Scientific"
__status__ = "Development"

from abc import ABC, abstractmethod

import numpy as np

from gscipy.time import Time, TimeList
from gscipy.trajectories import BaseTrajectoryCollection, Trajectory
from gscipy.signals import refill_gaps, central_diff, wavelet_filter
from gscipy.vectors import Vector, VectorsList

# TODO: evaluate if to remove __call__
# TODO: evaluate weather adding a transformation_function() method to the BaseOperator class
#  where the transformation logic can be stored as well as other methods required by it
# TODO: define set_params(), get_params() function in Base operator to get/set Operator args after initialization


def _reconstruct_trajectory(
    x: np.ndarray, y: np.ndarray, z: np.ndarray, time: np.ndarray
) -> Trajectory:
    vectors = VectorsList()
    times = TimeList()

    n = len(x)
    for i in range(n):
        r = {"x": x[i], "y": y[i], "z": z[i]}

        vector_i = Vector(**r)
        time_i = Time(time[i])

        vectors.append(vector_i)
        times.append(time_i)

    return Trajectory(vectors, times)


class BaseTrajectoryOperator(ABC):
    """Represents a mathematical function or operation transforming (mapping) a time-series of 3D vectors."""

    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        # Instantiate args and kwargs
        # self.args = args, self.kwargs = kwargs
        ...

    @abstractmethod
    def transform(
        self, trajectory: Trajectory
    ) -> Trajectory | BaseTrajectoryCollection:
        """Implement logic to transform trajectory."""
        ...

        # Template transformation logic for Trajectory:
        # x, y, z, time =  fun(trajectory, *self.args, **self.kwargs)
        # transformed_trajectory = _reconstruct_trajectory(x, y, z, time)
        # return transformed_trajectory

    def __call__(
        self, signal: Trajectory
    ) -> Trajectory | BaseTrajectoryCollection:
        return self.transform(signal)

    def __repr__(self) -> str:
        pars = ", ".join(
            f"{key}={value}" for key, value in self.__dict__.items()
        )
        return f"{self.__class__.__name__}({pars})"


class VectorFirstDiff(BaseTrajectoryOperator):
    def __init__(self, kernel_half_size: int = 1) -> None:
        self.kernel_half_size = kernel_half_size

    def transform(self, trajectory: Trajectory) -> Trajectory:
        # Transformation logic
        delta_x = central_diff(
            trajectory.x, kernel_half_size=self.kernel_half_size
        )
        delta_y = central_diff(
            trajectory.y, kernel_half_size=self.kernel_half_size
        )
        delta_z = central_diff(
            trajectory.z, kernel_half_size=self.kernel_half_size
        )
        time = trajectory.time

        dr = _reconstruct_trajectory(delta_x, delta_y, delta_z, time)

        return dr


class TimeFirstDiff(BaseTrajectoryOperator):
    def __init__(self, kernel_half_size: int = 1) -> None:
        self.kernel_half_size = kernel_half_size

    def transform(self, trajectory: Trajectory) -> Trajectory:
        # Transformation logic
        x = np.ones_like(trajectory.x)
        y = np.ones_like(trajectory.y)
        z = np.ones_like(trajectory.z)
        delta_t = central_diff(
            trajectory.time, kernel_half_size=self.kernel_half_size
        )

        delta_time = _reconstruct_trajectory(x, y, z, delta_t)

        return delta_time


class TimeDerivative(BaseTrajectoryOperator):
    def __init__(self, kernel_half_size: int = 1):
        self.kernel_half_size = kernel_half_size

    def transform(self, trajectory: Trajectory) -> Trajectory:
        delta_position = VectorFirstDiff(self.kernel_half_size)
        delta_time = TimeFirstDiff(self.kernel_half_size)

        delta_r = delta_position(trajectory)
        delta_t = delta_time(trajectory)

        # Transformation logic
        velocity_x = delta_r.x / delta_t.time
        velocity_y = delta_r.y / delta_t.time
        velocity_z = delta_r.z / delta_t.time
        time = delta_r.time

        velocity = _reconstruct_trajectory(
            velocity_x, velocity_y, velocity_z, time
        )

        return velocity


class WaveletFilter(BaseTrajectoryOperator):
    def __init__(
        self,
        wavelet_type: str,
        extension_mode: str,
        filtered_levels: tuple[int, ...],
    ) -> None:
        self.wavelet_type = wavelet_type
        self.extension_mode = extension_mode
        self.filtered_levels = filtered_levels

    def transform(self, trajectory: Trajectory) -> Trajectory:
        args = (self.wavelet_type, self.extension_mode, self.filtered_levels)

        x_filtered, _ = wavelet_filter(trajectory.x, *args)
        y_filtered, _ = wavelet_filter(trajectory.y, *args)
        z_filtered, _ = wavelet_filter(trajectory.z, *args)
        time = trajectory.time

        r_filtered = _reconstruct_trajectory(
            x_filtered, y_filtered, z_filtered, time
        )

        return r_filtered


class RefillGaps(BaseTrajectoryOperator):
    def __init__(self, max_lag: int = 10) -> None:
        self.max_lag = max_lag

    def transform(self, trajectory: Trajectory) -> Trajectory:
        x_refilled = refill_gaps(trajectory.x, self.max_lag)
        y_refilled = refill_gaps(trajectory.y, self.max_lag)
        z_refilled = refill_gaps(trajectory.z, self.max_lag)
        time = trajectory.time

        r_refilled = _reconstruct_trajectory(
            x_refilled, y_refilled, z_refilled, time
        )

        return r_refilled


class SequentialPipeline(BaseTrajectoryOperator):
    """Process the input trajectory sequentially using a list of Operator(s) instances."""

    def __init__(self, steps: list[BaseTrajectoryOperator]) -> None:
        self.steps = steps

    def transform(self, trajectory: Trajectory) -> Trajectory:
        transformed_trajectory = trajectory

        for operator in self.steps:
            transformed_trajectory = operator.transform(transformed_trajectory)

        return transformed_trajectory

    def __getitem__(self, index):
        """Get single operator (by indexing) or sub-pipeline (by slicing)."""
        if isinstance(index, slice):
            return SequentialPipeline(self.steps[index])
        elif isinstance(index, int):
            return self.steps[index]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(steps={[operator for operator in self.steps]})"


class ParallelPipeline(BaseTrajectoryOperator):
    """Process one input trajectory to many output trajectories using a list of Operator(s) instances."""

    def __init__(self, operators: list[BaseTrajectoryOperator]) -> None:
        self.operators = operators

    def transform(self, trajectory: Trajectory) -> list[Trajectory]:

        trajectories = [trajectory]

        for operator in self.operators:
            transformed_trajectory = operator.transform(trajectory)
            trajectories.append(transformed_trajectory)

        return trajectories

    def __getitem__(self, index):
        """Get single operator (by indexing) or sub-pipeline (by slicing)."""
        if isinstance(index, slice):
            return ParallelPipeline(self.operators[index])
        elif isinstance(index, int):
            return self.operators[index]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(operators={[operator for operator in self.operators]})"


class NthVectorDiff(SequentialPipeline):
    def __init__(self, kernel_half_size: int = 1, order: int = 1):

        pipeline_steps = [
            VectorFirstDiff(kernel_half_size) for _ in range(order)
        ]
        super().__init__(pipeline_steps)


class NthTimeDerivative(SequentialPipeline):
    def __init__(self, kernel_half_size: int = 1, order: int = 1):

        pipeline_steps = [
            TimeDerivative(kernel_half_size) for _ in range(order)
        ]
        super().__init__(pipeline_steps)


class DeltaR(NthVectorDiff):
    def __init__(self, kernel_half_size: int = 1):
        super().__init__(kernel_half_size, order=1)


class DeltaDeltaR(NthVectorDiff):
    def __init__(self, kernel_half_size: int = 1):
        super().__init__(kernel_half_size, order=2)


class Velocity(NthTimeDerivative):
    def __init__(self, kernel_half_size: int = 1):
        super().__init__(kernel_half_size, order=1)


class Acceleration(NthTimeDerivative):
    def __init__(self, kernel_half_size: int = 1):
        super().__init__(kernel_half_size, order=2)
